@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2001, 2020, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 2001, 2022, Oracle and/or its affiliates. All rights reserved.
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
  * This code is free software; you can redistribute it and/or modify it
@@ -27,6 +27,7 @@ package sun.nio.ch;
 
 import java.io.FileDescriptor;
 import java.io.IOException;
+import java.io.InterruptedIOException;
 import java.io.UncheckedIOException;
 import java.lang.invoke.MethodHandles;
 import java.lang.invoke.VarHandle;
@@ -66,14 +67,20 @@ import java.util.HashSet;
 import java.util.Map;
 import java.util.Objects;
 import java.util.Set;
+import java.util.concurrent.TimeUnit;
 import java.util.concurrent.locks.ReentrantLock;
 import java.util.function.Consumer;
 
+import jdk.internal.access.JavaNioAccess;
+import jdk.internal.access.SharedSecrets;
 import jdk.internal.ref.CleanerFactory;
 import sun.net.ResourceManager;
 import sun.net.ext.ExtendedSocketOptions;
 import sun.net.util.IPAddressUtil;
 
+import static java.util.concurrent.TimeUnit.MILLISECONDS;
+import static java.util.concurrent.TimeUnit.NANOSECONDS;
+
 /**
  * An implementation of DatagramChannels.
  */
@@ -85,6 +92,8 @@ class DatagramChannelImpl
     // Used to make native read and write calls
     private static final NativeDispatcher nd = new DatagramDispatcher();
 
+    private static final JavaNioAccess NIO_ACCESS = SharedSecrets.getJavaNioAccess();
+
     // true if interruptible (can be false to emulate legacy DatagramSocket)
     private final boolean interruptible;
 
@@ -159,6 +168,13 @@ class DatagramChannelImpl
     // set true/false when socket is already bound and SO_REUSEADDR is emulated
     private boolean isReuseAddress;
 
+    // True if the channel's socket has been forced into non-blocking mode
+    // by a virtual thread. It cannot be reset. When the channel is in
+    // blocking mode and the channel's socket is in non-blocking mode then
+    // operations that don't complete immediately will poll the socket and
+    // preserve the semantics of blocking operations.
+    private volatile boolean forcedNonBlocking;
+
     // -- End of fields protected by stateLock
 
 
@@ -470,6 +486,31 @@ class DatagramChannelImpl
         return DefaultOptionsHolder.defaultOptions;
     }
 
+    @Override
+    public void park(int event, long nanos) throws IOException {
+        Thread thread = Thread.currentThread();
+        if (thread.isVirtual()) {
+            Poller.poll(getFDVal(), event, nanos, this::isOpen);
+            // DatagramSocket throws when virtual thread interrupted
+            if (!interruptible && thread.isInterrupted()) {
+                throw new InterruptedIOException();
+            }
+        } else {
+            long millis;
+            if (nanos == 0) {
+                millis = -1;
+            } else {
+                millis = NANOSECONDS.toMillis(nanos);
+                if (nanos > MILLISECONDS.toNanos(millis)) {
+                    // Round up any excess nanos to the nearest millisecond to
+                    // avoid parking for less than requested.
+                    millis++;
+                }
+            }
+            Net.poll(getFD(), event, millis);
+        }
+    }
+
     /**
      * Marks the beginning of a read operation that might block.
      *
@@ -535,7 +576,9 @@ class DatagramChannelImpl
             SocketAddress sender = null;
             try {
                 SocketAddress remote = beginRead(blocking, false);
+                configureSocketNonBlockingIfVirtualThread();
                 boolean connected = (remote != null);
+                @SuppressWarnings("removal")
                 SecurityManager sm = System.getSecurityManager();
                 if (connected || (sm == null)) {
                     // connected or no security manager
@@ -571,21 +614,17 @@ class DatagramChannelImpl
      * manager.
      */
     private SocketAddress untrustedReceive(ByteBuffer dst) throws IOException {
+        @SuppressWarnings("removal")
         SecurityManager sm = System.getSecurityManager();
         assert readLock.isHeldByCurrentThread()
                 && sm != null && remoteAddress == null;
 
-        ByteBuffer bb = Util.getTemporaryDirectBuffer(dst.remaining());
-        try {
-            boolean blocking = isBlocking();
-            for (;;) {
-                int n = receive(bb, false);
-                if (blocking) {
-                    while (IOStatus.okayToRetry(n) && isOpen()) {
-                        park(Net.POLLIN);
-                        n = receive(bb, false);
-                    }
-                }
+        boolean blocking = isBlocking();
+        for (;;) {
+            int n;
+            ByteBuffer bb = Util.getTemporaryDirectBuffer(dst.remaining());
+            try {
+                n = receive(bb, false);
                 if (n >= 0) {
                     // sender address is in socket address buffer
                     InetSocketAddress isa = sourceSocketAddress();
@@ -596,14 +635,17 @@ class DatagramChannelImpl
                         return isa;
                     } catch (SecurityException se) {
                         // ignore datagram
-                        bb.clear();
                     }
-                } else {
-                    return null;
                 }
+            } finally {
+                Util.releaseTemporaryDirectBuffer(bb);
+            }
+
+            if (blocking && IOStatus.okayToRetry(n) && isOpen()) {
+                park(Net.POLLIN);
+            } else {
+                return null;
             }
-        } finally {
-            Util.releaseTemporaryDirectBuffer(bb);
         }
     }
 
@@ -622,6 +664,7 @@ class DatagramChannelImpl
             ensureOpen();
             if (!isBlocking())
                 throw new IllegalBlockingModeException();
+            @SuppressWarnings("removal")
             SecurityManager sm = System.getSecurityManager();
             boolean connected = isConnected();
             SocketAddress sender;
@@ -659,6 +702,7 @@ class DatagramChannelImpl
         SocketAddress sender = null;
         try {
             SocketAddress remote = beginRead(true, false);
+            configureSocketNonBlockingIfVirtualThread();
             boolean connected = (remote != null);
             int n = receive(dst, connected);
             while (IOStatus.okayToRetry(n) && isOpen()) {
@@ -716,6 +760,10 @@ class DatagramChannelImpl
         }
     }
 
+    /**
+     * Receives a datagram into the buffer.
+     * @param connected true if the channel is connected
+     */
     private int receive(ByteBuffer dst, boolean connected) throws IOException {
         int pos = dst.position();
         int lim = dst.limit();
@@ -744,13 +792,18 @@ class DatagramChannelImpl
                                         boolean connected)
         throws IOException
     {
-        int n = receive0(fd,
-                         ((DirectBuffer)bb).address() + pos, rem,
-                         sourceSockAddr.address(),
-                         connected);
-        if (n > 0)
-            bb.position(pos + n);
-        return n;
+        NIO_ACCESS.acquireSession(bb);
+        try {
+            int n = receive0(fd,
+                            ((DirectBuffer)bb).address() + pos, rem,
+                            sourceSockAddr.address(),
+                            connected);
+            if (n > 0)
+                bb.position(pos + n);
+            return n;
+        } finally {
+            NIO_ACCESS.releaseSession(bb);
+        }
     }
 
     /**
@@ -786,6 +839,7 @@ class DatagramChannelImpl
             boolean completed = false;
             try {
                 SocketAddress remote = beginWrite(blocking, false);
+                configureSocketNonBlockingIfVirtualThread();
                 if (remote != null) {
                     // connected
                     if (!target.equals(remote)) {
@@ -801,6 +855,7 @@ class DatagramChannelImpl
                     completed = (n > 0);
                 } else {
                     // not connected
+                    @SuppressWarnings("removal")
                     SecurityManager sm = System.getSecurityManager();
                     InetAddress ia = isa.getAddress();
                     if (sm != null) {
@@ -892,6 +947,7 @@ class DatagramChannelImpl
         int rem = (pos <= lim ? lim - pos : 0);
 
         int written;
+        NIO_ACCESS.acquireSession(bb);
         try {
             int addressLen = targetSocketAddress(target);
             written = send0(fd, ((DirectBuffer)bb).address() + pos, rem,
@@ -900,6 +956,8 @@ class DatagramChannelImpl
             if (isConnected())
                 throw pue;
             written = rem;
+        } finally {
+            NIO_ACCESS.releaseSession(bb);
         }
         if (written > 0)
             bb.position(pos + written);
@@ -933,6 +991,7 @@ class DatagramChannelImpl
             int n = 0;
             try {
                 beginRead(blocking, true);
+                configureSocketNonBlockingIfVirtualThread();
                 n = IOUtil.read(fd, buf, -1, nd);
                 if (blocking) {
                     while (IOStatus.okayToRetry(n) && isOpen()) {
@@ -962,6 +1021,7 @@ class DatagramChannelImpl
             long n = 0;
             try {
                 beginRead(blocking, true);
+                configureSocketNonBlockingIfVirtualThread();
                 n = IOUtil.read(fd, dsts, offset, length, nd);
                 if (blocking) {
                     while (IOStatus.okayToRetry(n)  && isOpen()) {
@@ -1044,6 +1104,7 @@ class DatagramChannelImpl
             int n = 0;
             try {
                 beginWrite(blocking, true);
+                configureSocketNonBlockingIfVirtualThread();
                 n = IOUtil.write(fd, buf, -1, nd);
                 if (blocking) {
                     while (IOStatus.okayToRetry(n) && isOpen()) {
@@ -1073,6 +1134,7 @@ class DatagramChannelImpl
             long n = 0;
             try {
                 beginWrite(blocking, true);
+                configureSocketNonBlockingIfVirtualThread();
                 n = IOUtil.write(fd, srcs, offset, length, nd);
                 if (blocking) {
                     while (IOStatus.okayToRetry(n) && isOpen()) {
@@ -1112,21 +1174,21 @@ class DatagramChannelImpl
         assert readLock.isHeldByCurrentThread() || writeLock.isHeldByCurrentThread();
         synchronized (stateLock) {
             ensureOpen();
-            IOUtil.configureBlocking(fd, block);
+            // do nothing if virtual thread has forced the socket to be non-blocking
+            if (!forcedNonBlocking) {
+                IOUtil.configureBlocking(fd, block);
+            }
         }
     }
 
     /**
-     * Adjusts the blocking mode if the channel is open. readLock or writeLock
-     * must already be held.
-     *
-     * @return {@code true} if the blocking mode was adjusted, {@code false} if
-     *         the blocking mode was not adjusted because the channel is closed
+     * Attempts to adjust the blocking mode if the channel is open.
+     * @return {@code true} if the blocking mode was adjusted
      */
     private boolean tryLockedConfigureBlocking(boolean block) throws IOException {
         assert readLock.isHeldByCurrentThread() || writeLock.isHeldByCurrentThread();
         synchronized (stateLock) {
-            if (isOpen()) {
+            if (!forcedNonBlocking && isOpen()) {
                 IOUtil.configureBlocking(fd, block);
                 return true;
             } else {
@@ -1135,6 +1197,21 @@ class DatagramChannelImpl
         }
     }
 
+    /**
+     * Ensures that the socket is configured non-blocking when on a virtual thread.
+     * @throws IOException if there is an I/O error changing the blocking mode
+     */
+    private void configureSocketNonBlockingIfVirtualThread() throws IOException {
+        assert readLock.isHeldByCurrentThread() || writeLock.isHeldByCurrentThread();
+        if (!forcedNonBlocking && Thread.currentThread().isVirtual()) {
+            synchronized (stateLock) {
+                ensureOpen();
+                IOUtil.configureBlocking(fd, false);
+                forcedNonBlocking = true;
+            }
+        }
+    }
+
     InetSocketAddress localAddress() {
         synchronized (stateLock) {
             return localAddress;
@@ -1182,6 +1259,7 @@ class DatagramChannelImpl
         } else {
             isa = Net.checkAddress(local, family);
         }
+        @SuppressWarnings("removal")
         SecurityManager sm = System.getSecurityManager();
         if (sm != null)
             sm.checkListen(isa.getPort());
@@ -1210,6 +1288,7 @@ class DatagramChannelImpl
      */
     DatagramChannel connect(SocketAddress sa, boolean check) throws IOException {
         InetSocketAddress isa = Net.checkAddress(sa, family);
+        @SuppressWarnings("removal")
         SecurityManager sm = System.getSecurityManager();
         if (sm != null) {
             InetAddress ia = isa.getAddress();
@@ -1257,7 +1336,7 @@ class DatagramChannelImpl
                     // flush any packets already received.
                     boolean blocking = isBlocking();
                     if (blocking) {
-                        IOUtil.configureBlocking(fd, false);
+                        lockedConfigureBlocking(false);
                     }
                     try {
                         ByteBuffer buf = ByteBuffer.allocate(100);
@@ -1266,7 +1345,7 @@ class DatagramChannelImpl
                         }
                     } finally {
                         if (blocking) {
-                            IOUtil.configureBlocking(fd, true);
+                            tryLockedConfigureBlocking(true);
                         }
                     }
                 }
@@ -1369,7 +1448,7 @@ class DatagramChannelImpl
             }
 
             // copy the blocking mode
-            if (!isBlocking()) {
+            if (!isBlocking() || forcedNonBlocking) {
                 IOUtil.configureBlocking(newfd, false);
             }
 
@@ -1425,6 +1504,7 @@ class DatagramChannelImpl
     /**
      * Defines static methods to access AbstractSelectableChannel non-public members.
      */
+    @SuppressWarnings("removal")
     private static class AbstractSelectableChannels {
         private static final Method FOREACH;
         static {
@@ -1481,6 +1561,7 @@ class DatagramChannelImpl
                 throw new IllegalArgumentException("Source address is different type to group");
         }
 
+        @SuppressWarnings("removal")
         SecurityManager sm = System.getSecurityManager();
         if (sm != null)
             sm.checkMulticast(group);
@@ -1724,11 +1805,18 @@ class DatagramChannelImpl
                 long reader = readerThread;
                 long writer = writerThread;
                 if (reader != 0 || writer != 0) {
-                    nd.preClose(fd);
-                    if (reader != 0)
-                        NativeThread.signal(reader);
-                    if (writer != 0)
-                        NativeThread.signal(writer);
+                    if (NativeThread.isVirtualThread(reader)
+                            || NativeThread.isVirtualThread(writer)) {
+                        Poller.stopPoll(fdVal);
+                    }
+                    if (NativeThread.isNativeThread(reader)
+                            || NativeThread.isNativeThread(writer)) {
+                        nd.preClose(fd);
+                        if (NativeThread.isNativeThread(reader))
+                            NativeThread.signal(reader);
+                        if (NativeThread.isNativeThread(writer))
+                            NativeThread.signal(writer);
+                    }
                 }
             }
         }
